/* Skull stripping filter to preprocess T1-weighted MR images for SSM construction*/

#include "time.h"
#include "itkImageFileReader.h"
#include "itkImageFileWriter.h"
#include "itkMaskImageFilter.h"
#include "itkStripTsImageFilter.h"
#include "TypeDefinitions.h"
#include <iostream>
#include <fstream>
#include <unistd.h>

void print_help()
{
      std::cerr << std::endl;
      std::cerr << "Perform atlas based skull stripping on patient images" << std::endl;
      std::cerr << "Usage: SkullStripping " << std::endl; 
      std::cerr << " -n number of input images " << std::endl;
      std::cerr << " -a Input atlas image " << std::endl;
      std::cerr << " -f Input patient images " << std::endl;
      std::cerr << " -s startIndex (default: 1)"  << std::endl;
      std::cerr << " -m Input atlas mask image " << std::endl;
      std::cerr << " -o Patient mask pattern (brainMask%02d.nii) " << std::endl;
      std::cerr << " -b Patient brain extracted image  (brainExt_%02d.nii) " << std::endl;
      std::cerr << " SkullStripping -n 5 -s 1 -a Atlas.nii -m AtlasMask.nii -o PatientMask%02d.nii -b PatientBrain%02d.nii" << std::endl;
      std::cerr << std::endl;
}

int main( int argc, char* argv[] )
{
 int c=0;
 int rOpts=0;
 char inputSeriesPattern[1000];
 char inputAtlas[1000];
 char inputAtlasMask[1000];
 char outMaskPattern[1000]="brainMask%02d.nii";
 char BEimage[1000]="brainExt_%02d.nii";
 unsigned int st_index = 1;
 unsigned int nb_train=0;
 
 while ((c = getopt (argc, argv, "s:n:f:a:m:o:b")) != -1)
    switch (c)
      {
      case 'n':
	nb_train = atoi(optarg);
	++rOpts;
	break;
      case 'f':
	sprintf(inputSeriesPattern,"%s",optarg);
	++rOpts;
	break;
      case 'a':
	sprintf(inputAtlas,"%s",optarg);
	break;
      case 'm':
	sprintf(inputAtlasMask,"%s",optarg);
	break;
      case 's':
	st_index = atoi(optarg);
	break;
      case 'o':
	sprintf(outMaskPattern,"%s",optarg);
	break;
       case 'b':
	sprintf(BEimage,"%s",optarg);
	break;
      case '?':
	fprintf (stderr, "Unknown option %s.\n", optarg);
	return 1;
      default:
	abort ();
     }
 if (rOpts != 2)
   {
     print_help();
     return 1;
   }

  double startTime = time(0);

  typedef itk::Image<short, 3> AtlasImageType;
  typedef itk::Image<unsigned char, 3> AtlasLabelType;


  // image reading
  typedef itk::ImageFileReader<ImageType> ReaderType;
  typedef itk::ImageFileReader<AtlasImageType> AtlasReaderType;
  typedef itk::ImageFileReader<AtlasLabelType> LabelReaderType;
  ReaderType::Pointer reader = ReaderType::New();
  AtlasReaderType::Pointer atlasReader = AtlasReaderType::New();
  LabelReaderType::Pointer labelReader = LabelReaderType::New();


   itk::NumericSeriesFileNames::Pointer inputFileNamesCreator =
                                       itk::NumericSeriesFileNames::New();
  std::vector<ImageType::Pointer> trainingImages( nb_train );

  inputFileNamesCreator->SetStartIndex( st_index );
  inputFileNamesCreator->SetEndIndex( st_index + nb_train  );
  inputFileNamesCreator->SetSeriesFormat( inputSeriesPattern );
  const std::vector<std::string> & patientFileNames =
                                   inputFileNamesCreator->GetFileNames();

  std::vector<ReaderType::Pointer> readerVector(nb_train);

  for ( unsigned int k = 0; k < nb_train; k++ )
    {
      readerVector[k] = ReaderType::New();
      readerVector[k]->SetFileName( patientFileNames[k].c_str() );
      readerVector[k]->Update();
      trainingImages[k] = readerVector[k]->GetOutput();
    }

  atlasReader->SetFileName(inputAtlas);
  labelReader->SetFileName(inputAtlasMask);

  try
    {
    atlasReader->Update();
    labelReader->Update();
    }
  catch ( itk::ExceptionObject &exception )
    {
    std::cerr << "Exception caught ! " << std::endl;
    std::cerr << exception << std::endl;

    return EXIT_FAILURE;
    }

  for (unsigned int k=0; k<nb_train; k++)
    {
  // perform skull-stripping using stripTsImageFilter
  std::cout << std::endl << "Performing skull-stripping for patient: " << k << std::endl;

  // set up skull-stripping filter
  typedef itk::StripTsImageFilter<ImageType, AtlasImageType, AtlasLabelType> StripTsFilterType;
  StripTsFilterType::Pointer stripTsFilter = StripTsFilterType::New();

  // set the required inputs for the stripTsImageFilter
  stripTsFilter->SetInput( trainingImages[k] );
  stripTsFilter->SetAtlasImage( atlasReader->GetOutput() );
  stripTsFilter->SetAtlasBrainMask( labelReader->GetOutput() );

  try
    {
    stripTsFilter->Update();
    }
  catch ( itk::ExceptionObject &exception )
    {
    std::cerr << "Exception caught ! " << std::endl;
    std::cerr << exception << std::endl;

    return EXIT_FAILURE;
    }


  // mask the patient image using the output generated from the stripTsImageFilter as mask
  typedef itk::MaskImageFilter<ImageType, AtlasLabelType, ImageType> MaskFilterType;
  MaskFilterType::Pointer maskFilter = MaskFilterType::New();

  maskFilter->SetInput1( trainingImages[k] );
  maskFilter->SetInput2( stripTsFilter->GetOutput() );

  try
    {
    maskFilter->Update();
    }
  catch ( itk::ExceptionObject &exception )
    {
    std::cerr << "Exception caught ! " << std::endl;
    std::cerr << exception << std::endl;

    return EXIT_FAILURE;
    }

  // Create mask filenames
  itk::NumericSeriesFileNames::Pointer maskNamesCreator =
                                       itk::NumericSeriesFileNames::New();  
  maskNamesCreator->SetStartIndex( st_index );
  maskNamesCreator->SetEndIndex( st_index + nb_train  );
  maskNamesCreator->SetSeriesFormat( outMaskPattern);
  const std::vector<std::string> & maskFileNames =
                                  maskNamesCreator->GetFileNames();

 itk::NumericSeriesFileNames::Pointer outputFileNamesCreator =
                                       itk::NumericSeriesFileNames::New();  
 // Create brain extracted image filenames
  outputFileNamesCreator->SetStartIndex( st_index );
  outputFileNamesCreator->SetEndIndex( st_index + nb_train  );
  outputFileNamesCreator->SetSeriesFormat(BEimage);
  const std::vector<std::string> & BEimgNames =
                                   outputFileNamesCreator->GetFileNames();

  // write mask and masked patient image
  typedef itk::ImageFileWriter<AtlasLabelType> MaskWriterType;
  MaskWriterType::Pointer maskWriter = MaskWriterType::New();
  maskWriter->SetInput( stripTsFilter->GetOutput() );
  maskWriter->SetFileName((char *)maskFileNames[k].c_str());

  typedef itk::ImageFileWriter<ImageType> ImageWriterType;
  ImageWriterType::Pointer imageWriter = ImageWriterType::New();
  imageWriter->SetInput( maskFilter->GetOutput() );
  imageWriter->SetFileName((char *)BEimgNames[k].c_str());

  try
    {
    maskWriter->Update();
    imageWriter->Update();
    }
  catch ( itk::ExceptionObject &exception )
    {
    std::cerr << "Exception caught ! " << std::endl;
    std::cerr << exception << std::endl;

    return EXIT_FAILURE;
    }


  double endTime = time(0);
  std::cout << "Total computation time: " << endTime-startTime << "seconds " << std::endl;

    }

  return EXIT_SUCCESS;
}
