#include "itkImage.h"
#include "itkImageFileReader.h"
#include "itkImageFileWriter.h"


template<typename TImageType>
void ReadWriteFile(std::string inputFilename, std::string outputFilename, typename TImageType::Pointer image)
{
  typedef itk::ImageFileReader<TImageType> ReaderType;
  typename ReaderType::Pointer reader = ReaderType::New();
 
  reader->SetFileName(inputFilename);
  reader->Update();
 
  image->Graft(reader->GetOutput());

  typedef itk::ImageFileWriter<TImageType> WriterType;
  typename WriterType::Pointer writer = WriterType::New();

  writer->SetInput(image);
  writer->SetFileName(outputFilename);
  writer->Update();

}

int main( int argc, char * argv[] )
{
  if( argc < 3 )
    {
    std::cerr << "Usage: " << std::endl;
    std::cerr << argv[0] << " inputImage.mhd outputImage.hdr" << std::endl;
    return EXIT_FAILURE;
    }
  std::string inputFilename = argv[1];
  std::string outputFilename = argv[2];

  typedef itk::ImageIOBase::IOComponentType ScalarPixelType;
 
  itk::ImageIOBase::Pointer imageIO =
        itk::ImageIOFactory::CreateImageIO(
            inputFilename.c_str(), itk::ImageIOFactory::ReadMode);
 
  imageIO->SetFileName(inputFilename);
  imageIO->ReadImageInformation();
  const ScalarPixelType pixelType = imageIO->GetComponentType();
  //std::cout << "Pixel Type is " << imageIO->GetComponentTypeAsString(pixelType) // 'double'
  //          << std::endl;
  const size_t numDimensions =  imageIO->GetNumberOfDimensions();
  //std::cout << "numDimensions: " << numDimensions << std::endl; // '2'
 
  //std::cout << "component size: " << imageIO->GetComponentSize() << std::endl; // '8'
  //std::cout << "pixel type (string): " << imageIO->GetPixelTypeAsString(imageIO->GetPixelType()) << std::endl; // 'vector'
  //std::cout << "pixel type: " << imageIO->GetPixelType() << std::endl; // '5'
 
  switch (pixelType)
  {
  case itk::ImageIOBase::UCHAR:
    {
    typedef itk::Image<unsigned char, 3> ImageType;
    ImageType::Pointer image = ImageType::New();
    ReadWriteFile<ImageType>(inputFilename, outputFilename, image);
    }
    break;
  case itk::ImageIOBase::CHAR:
    {
    typedef itk::Image<char, 3> ImageType;
    ImageType::Pointer image = ImageType::New();
    ReadWriteFile<ImageType>(inputFilename, outputFilename, image);
    }
    break;

  case itk::ImageIOBase::USHORT:
    {
    typedef itk::Image<unsigned short, 3> ImageType;
    ImageType::Pointer image = ImageType::New();
    ReadWriteFile<ImageType>(inputFilename, outputFilename, image);
    }
    break;
  case itk::ImageIOBase::SHORT:
    {
    typedef itk::Image<short, 3> ImageType;
    ImageType::Pointer image = ImageType::New();
    ReadWriteFile<ImageType>(inputFilename, outputFilename, image);
    }
    break;
  case itk::ImageIOBase::FLOAT:
    {
    typedef itk::Image<float, 3> ImageType;
    ImageType::Pointer image = ImageType::New();
    ReadWriteFile<ImageType>(inputFilename, outputFilename, image);
    }
    break;
 
  default:
    std::cerr << "Pixel Type ("
	      << imageIO->GetComponentTypeAsString(pixelType)
	      << ") not supported. Exiting." << std::endl;
    return -1;
  }
  

  return EXIT_SUCCESS;
}

