/*
Author: Ali Gooya <ali dot gooya at gmail dot com>
*/
#include "TypeDefinitions.h"
#include <iostream>
#include <fstream>
#include <unistd.h>


void GetVNLPoint(PointSetType * points, int index, VNLPointType *vnlp) //concatenates co-ordinates with feature values
{
  VectorType f;
  points->GetPointData(index,&f); 
  PointType p;
  points->GetPoint(index,&p);
  
  for (int k = 0; k < Dimension; k++)
    {
      (*vnlp)[k] = p[k];
    }
  for (int k = Dimension; k < GMMPointDimension; k++)
    {
      (*vnlp)[k] = f[k - Dimension];
    }
}


void WriteMesh(PointSetType * points,  char * name)
{
  std::ofstream myfile;
  myfile.open(name);
  //std::cout << "writing points" << std::endl;

  PointsIteratorType it = points->GetPoints()->Begin();
  while (it != points->GetPoints()->End())
    {
      if ( it != points->GetPoints()->Begin())
	myfile << std::endl;
      VNLPointType vnlp;
      GetVNLPoint(points, it.Index() , &vnlp);
      myfile << vnlp ;
      ++it;
    }
  //myfile << std::ends;
  myfile.close();

}

void print_help()
{
      std::cerr << std::endl;
      std::cerr << "Constructs xyz/xyz-l point sets from segmented images" << std::endl;
      std::cerr << "Usage: imagesToPointSets " << std::endl; 
      std::cerr << " -n number of input images " << std::endl;
      std::cerr << " -f trainingFilePattern " << std::endl;
      std::cerr << "Options: " << std::endl;
      std::cerr << " -s startIndex (default: 1)"  << std::endl;
      std::cerr << " -T levelSetThick (default: 1.5mm )" << std::endl;
      std::cerr << " -l lowerThreshold (default: .01) " << std::endl;
      std::cerr << " -u upperThreshold (default: inf) " << std::endl;
      std::cerr << " -o outputFilePattern (MESH_%02d.txt) " << std::endl;
      std::cerr << " MeshSampleImages -n 5 -s 1 -f IBSR_%02d_seg_caudate_RAI.mhd" << std::endl;
      std::cerr << std::endl;
}

int main( int argc, char *argv[] )
{

  int c = 0;
  int rOpts = 0;
  unsigned int st_index = 1;
  unsigned int nb_train=0;
  char inputSeriesPattern[1000];
  char outputFilePattern[1000]="MESH_%02d.txt";
  float lowThr = 0.01;
  float upThr = itk::NumericTraits< FloatPixelType >::max();
  float nbThick = 1.5;

  while ((c = getopt (argc, argv, "s:n:f:T:l:u:o:")) != -1)
    switch (c)
      {
      case 'n':
	nb_train = atoi(optarg);
	++rOpts;
	break;
      case 'f':
	sprintf(inputSeriesPattern,"%s",optarg);
	++rOpts;
	break;
      case 's':
	st_index = atoi(optarg);
	break;
      case 'T':
	nbThick = atof(optarg);
       	break;
      case 'l':
	lowThr = atof(optarg);
       	break;
      case 'u':
	upThr = atof(optarg);
       	break;
      case 'o':
	sprintf(outputFilePattern,"%s",optarg);
	break;
      case '?':
	fprintf (stderr, "Unknown option %s.\n", optarg);
	return 1;
      default:
	abort ();
     }
  if (rOpts != 2)
    {
      print_help();
      return 1;
    }




  itk::NumericSeriesFileNames::Pointer inputFileNamesCreator =
                                       itk::NumericSeriesFileNames::New();
  std::vector<ImageType::Pointer> trainingImages( nb_train );

  inputFileNamesCreator->SetStartIndex( st_index );
  inputFileNamesCreator->SetEndIndex( st_index + nb_train  );
  inputFileNamesCreator->SetSeriesFormat( inputSeriesPattern );
  const std::vector<std::string> & shapeModeFileNames =
                                   inputFileNamesCreator->GetFileNames();

  std::vector<ReaderType::Pointer> readerVector(nb_train);

  for ( unsigned int k = 0; k < nb_train; k++ )
    {
       readerVector[k] = ReaderType::New();
      readerVector[k]->SetFileName( shapeModeFileNames[k].c_str() );
      readerVector[k]->Update();
      trainingImages[k] = readerVector[k]->GetOutput();
 
    }


  //Next, we make pointset

  //Something to keep the point sets
  PointSetVectorType pointSetVector( nb_train );

  std::vector<BinaryThresholdImageFilterType::Pointer> binThresholdFilterVector(nb_train);
  for ( unsigned int k = 0; k < nb_train; k++ )
    {
      binThresholdFilterVector[k] = BinaryThresholdImageFilterType::New();
      binThresholdFilterVector[k]->SetLowerThreshold(lowThr);
      binThresholdFilterVector[k]->SetUpperThreshold(upThr);
      binThresholdFilterVector[k]->SetInsideValue(itk::NumericTraits< PixelType >::One);
      binThresholdFilterVector[k]->SetOutsideValue(itk::NumericTraits< PixelType >::Zero);
      Cast2FloatImageFilterType::Pointer caster = Cast2FloatImageFilterType::New();
      caster->SetInput(trainingImages[k]);
      caster->Update();
      binThresholdFilterVector[k]->SetInput(caster->GetOutput());
      //binThresholdFilterVector[k]->SetInput(trainingImages[k]);
      binThresholdFilterVector[k]->Update();
    }


  std::vector<NarrowBandFilterType::Pointer> narrowBandPointSetFilter(nb_train);
   for ( unsigned int k = 0; k < nb_train; k++ )
    {
      narrowBandPointSetFilter[k] = NarrowBandFilterType::New();
      narrowBandPointSetFilter[k]->SetBandWidth( nbThick );
      narrowBandPointSetFilter[k]->SetInput( binThresholdFilterVector[k]->GetOutput() );
      narrowBandPointSetFilter[k]->Update();
      pointSetVector[k] = narrowBandPointSetFilter[k]->GetOutput();
     }
  
   //now write pointSets to files
  itk::NumericSeriesFileNames::Pointer outputFileNamesCreator =
                                       itk::NumericSeriesFileNames::New();  
  outputFileNamesCreator->SetStartIndex( st_index );
  outputFileNamesCreator->SetEndIndex( st_index + nb_train  );
  outputFileNamesCreator->SetSeriesFormat( outputFilePattern);
  const std::vector<std::string> & meshFileNames =
                                   outputFileNamesCreator->GetFileNames();

   for ( unsigned int k = 0; k < nb_train; k++ )
    {
      WriteMesh(pointSetVector[k], (char *)meshFileNames[k].c_str());
    }

  return EXIT_SUCCESS;
}

