/*
Author: Ali Gooya <ali dot gooya at gmail dot com>
*/
#include "stdafx.h"

#include "TypeDefinitions.h"
#include "math.h"
#include <iostream>
#include <fstream>
#include <time.h>
#include <unistd.h>
#include <itkNeighborhoodIterator.h>
#include <itkDerivativeOperator.h>
#include "interpolation.h"
#include <itkNiftiImageIO.h>
#include<itkNeighborhoodInnerProduct.h>
#include "itkConstNeighborhoodIterator.h"

#include <itkLinearInterpolateImageFunction.h>
#include <itkContinuousIndex.h>
#include <libgen.h>

#include "itkGeodesicActiveContourLevelSetImageFilter.h"
#include "itkCurvatureAnisotropicDiffusionImageFilter.h"
#include "itkFastMarchingImageFilter.h"
#include <itkReinitializeLevelSetImageFilter.h>
#include <itkCurvatureFlowImageFilter.h>
#include "itkBinaryThresholdImageFilter.h"
#include <itkBinaryMinMaxCurvatureFlowImageFilter.h>

#include <flann/flann.hpp> 
#include <itkGeodesicActiveContourWithPropagationLevelSetFunction.h>

//#include <flann/io/hdf5.h>
//#include <vnl/algo/vnl_symmetric_eigensystem.h>

using namespace alglib;

double computeDistance(std::vector< vnl_vector_fixed<float,Dimension> > & nps, vnl_vector_fixed<float,Dimension> & qp)
{
    double d = 0;
    int nn = nps.size();
    vnl_vector_fixed<float, Dimension> mean;
    mean.fill(0);
    for(int n=0; n < nn; n++) mean += nps[n];
    mean = (1/(float)nn)*mean;
    vnl_matrix<double> C(Dimension,Dimension);
    C.fill(0);
    for(int n=0; n < nn; n++) 
      {
	vnl_vector_fixed<float,Dimension> v = nps[n] - mean;
	vnl_vector_fixed<float,Dimension> dn = nps[n] - qp;
	double w = exp(-0.01*dn.magnitude());
	for (int i=0; i<Dimension; i++)
	  for (int j=0; j<Dimension; j++)
	    C(i,j) += (w*v(i)*v(j)/(double)nn);
      }
    vnl_svd< double > svd(C);
    vnl_vector_fixed<float,Dimension> qP = qp - mean;
    for (int i=0; i<Dimension; i++)
      {
	d += qP(i)*svd.U(i,Dimension-1);
      }
    return vnl_math_abs(d);
    //return d*d;
}


void SetVNLPoint(PointSetType * points, int index, VNLPointType *vnlp) //concatenates co-ordinates with feature values
{
  VectorType f;

  PointType p;
  
  for (int k = 0; k < Dimension; k++)
    {
       p[k] = (*vnlp)[k];
    }
  for (int k = Dimension; k < GMMPointDimension; k++)
    {
       f[k - Dimension] = (*vnlp)[k];
    }

  points->GetPoints()->InsertElement(index,p);
  points->GetPointData()->InsertElement(index,f);   
  //points->GetPointData()SetPoint(index,p);
  
}

void print_help()
{
  std::cerr << std::endl;
  std::cerr << "\tReads a point set file and makes a signed distance map." << std::endl;
  std::cerr << "\tUsage: pointSetToSignedDistanceUsingGAC " << std::endl;
  std::cerr << "\t -i inputPointSet ( x y z co-ordinates at each line)." << std::endl;
  std::cerr << "\tOptions:" << std::endl;
  std::cerr << "\t -t templateImage " << std::endl;
  std::cerr << "\t -o outputImage (default: <inputPointSet>.nii) " << std::endl;
  std::cerr << "\t -x -y -z sampling distances (ignored if templateimage is provided, default: 1.0 1.0 1.0)" << std::endl;
  std::cerr << "\t -d additional proportional width used for framing (default: .2)" << std::endl;
  std::cerr << "\t -k curvature weight (default: 2.0) " << std::endl;
  std::cerr << "\t -g number of GAC PDE iterations (default: 1000) " << std::endl;
  std::cerr << "\t -T thickness of narrowband for svd made distance (default: 10) " << std::endl;
  std::cerr << "\t -N number of nearest neighbors (default: 20) " << std::endl;
  std::cerr << "\t -p propagation weight (default: 100, with neg in/ pos out) " << std::endl;
  std::cerr << "\t -c number of distance filtering iterations (default: 20) " << std::endl;
  std::cerr << "\t -m number of min flow (hole filling used for large convex structures) iterations (default: 0) " << std::endl;
  std::cerr << "\t -G set if intremediate images needed (default: 0) " << std::endl;
  std::cerr << std::endl;

}


int main( int argc, char *argv[] )
{
 
  int rOpts = 0;
  int c;
  char inputMeshName[1000];
  char templateName[1000];
  char outputImageName[1000];
  int oNameProvided = 0;
  int templateProvided = 0;
  float dX=1.0,dY=1.0,dZ=1.0;
  float d = .2; //for numerical evaluations
  float gIter = 1000;
  float curveWeight = 1.0;
  float T = 10.0;
  int nn = 20;
  float propWeight = 100;
  int mIter = 0;
  int cIter = 20; 
  int debug = 0;
  
  while ((c = getopt (argc, argv, "i:t:o:x:y:z:d:k:g:T:N:p:m:c:G:")) != -1)
    switch (c)
      {
      case 'i':
	sprintf(inputMeshName,"%s",optarg);
	++rOpts;
	break;
      case 't':
	sprintf(templateName,"%s",optarg);
	templateProvided = 1;
	break;
      case 'o':
	sprintf(outputImageName,"%s",optarg);
	oNameProvided = 1;
	break;
      case 'x':
	dX = atof(optarg);
	break;
      case 'y':
	dY = atof(optarg);
	break;
      case 'z':
	dZ = atof(optarg);
	break;
      case 'd':
	d = atof(optarg);
	break;
      case 'g':
	gIter = atoi(optarg);
	break;
      case 'k':
	curveWeight = atof(optarg);
	break;
      case 'T':
	T = atof(optarg);
	break;
      case 'p':
	propWeight = atof(optarg);
	break;
      case 'N':
	nn = atoi(optarg);
	break;
      case 'm':
	mIter = atoi(optarg);
	break;
      case 'c':
	cIter = atoi(optarg);
	break;
      case 'G':
	debug = atoi(optarg);
	break;
      case '?':
	fprintf (stderr, "Unknown option %s.\n", optarg);
	return 1;
      default:
	abort ();
      }
  
 

  if (rOpts != 1 )
    {
      print_help();
      return 1;
    }
  // a file name for log file
  char logFileName[1000];
  std::string str= basename(inputMeshName);
  unsigned found = str.find_last_of(".");
  char bname[1000];
  sprintf(bname,"%s",(str.substr(0,found)).c_str());
  char currdir[1000];
  char * unused=getcwd(currdir,1000);
  sprintf(logFileName,"%s/%s%s",currdir,bname,"_logfile.txt");
  std::ofstream logfile;
  logfile.open(logFileName);
  time_t rawtime;
  struct tm * timeinfo;
  time ( &rawtime );
  timeinfo = localtime ( &rawtime ); 
  logfile << asctime(timeinfo) << std::endl;
  for (int i=0; i< argc; i++)
    {
      logfile << argv[i] << " ";
    }
  logfile << std::endl;

  
  std::fstream meshfile;
  meshfile.open(inputMeshName);

  typedef vnl_vector_fixed < float, Dimension > PointType;
  typedef std::vector< PointType > PointSetType;
  PointSetType pointSet;
  //PointSetType::Pointer  pointSet = PointSetType::New();
  //read mesh
  int cnt = 0;
  PointType minCorner;
  for (int i=0; i<Dimension; i++) minCorner[i]=1e32;
  PointType maxCorner;
  for (int i=0; i<Dimension; i++) maxCorner[i]=-1e32;
  while( !meshfile.eof() )
    {
      PointType p;
      meshfile >> p;
      for (int i=0; i<Dimension; i++){if(p[i] < minCorner[i]) minCorner[i] = p[i];}
      for (int i=0; i<Dimension; i++){if(p[i] > maxCorner[i]) maxCorner[i] = p[i];}
      pointSet.push_back(p);
      ++cnt;
      //SetVNLPoint( pointSet, cnt++, &vnlp);
      meshfile.ignore(1000000,'\n');
     }
  meshfile.close();

  ImageType::SizeType size;
  ImageType::SpacingType spacing;
  itk::Point<double,Dimension> origin;
  ImageType::DirectionType direction;
  
  itk::Point<double,Dimension> span;
  itk::Point<double,Dimension> startpoint;
  itk::Point<double,Dimension> endpoint;

  if ( templateProvided )
    {
      ReaderType::Pointer reader = ReaderType::New();
      reader->SetFileName(templateName);
      reader->Update();

      size = reader->GetOutput()->GetLargestPossibleRegion().GetSize();
      spacing = reader->GetOutput()->GetSpacing();
      origin = reader->GetOutput()->GetOrigin();
      direction = reader->GetOutput()->GetDirection();

      ImageType::IndexType ind ;
      for(int i=0; i<Dimension; i++) ind[i]=0;
 
      reader->GetOutput()->TransformIndexToPhysicalPoint( ind, startpoint );

      ind[0] = size[0] -1; ind[1] = 0; ind[2] =0;
      reader->GetOutput()->TransformIndexToPhysicalPoint( ind, endpoint );
      span[0] = endpoint[0] - startpoint[0];
      ind[0] = 0; ind[1] = size[1]-1; ind[2] =0;
      reader->GetOutput()->TransformIndexToPhysicalPoint( ind, endpoint );
      span[1] = endpoint[1] - startpoint[1];
      ind[0] = 0; ind[1] = 0; ind[2] =size[2]-1;
      reader->GetOutput()->TransformIndexToPhysicalPoint( ind, endpoint );
      span[2] = endpoint[2] - startpoint[2];
      
    }
  else
    {
      spacing[0] = dX;
      spacing[1] = dY;
      spacing[2] = dZ;
  
      for(int i=0; i< Dimension; i++) size[i] = int((1.0+d)*(maxCorner[i] - minCorner[i])/spacing[i]);
      for(int i=0; i< Dimension; i++) origin[i] = minCorner[i] - d*(maxCorner[i] - minCorner[i])/2.0;
      //for(int i=0; i< Dimension; i++) origin[i] = minCorner[i];
     
      //for(int i=0; i< Dimension; i++) span[i] = (1.0+d)*(maxCorner[i] - minCorner[i]);
      //for(int i=0; i< Dimension; i++) startpoint[i] = origin[i];
 
      direction[0][0]=1;direction[0][1]=0;direction[0][2]=0;
      direction[1][0]=0;direction[1][1]=1;direction[1][2]=0;
      direction[2][0]=0;direction[2][1]=0;direction[2][2]=1;

       
    }

 //getting back to work
  WriterType::Pointer writer = WriterType::New();
 
  int rows = cnt;
  int cols = Dimension;
  flann::Matrix<float> points(new float[rows*cols],rows,cols);
  for (int i=0; i< cnt; i++)
    {
	for (int j=0; j< Dimension; j++) 
	  {
	    points[i][j] = pointSet[i][j];
	  }
   }

  typedef itk::Image<float, Dimension>  OutputImageType;
  OutputImageType::Pointer image = OutputImageType::New();
  image->SetRegions(size);
  image->Allocate();
  image->SetOrigin(origin);
  image->SetSpacing(spacing );
  image->SetDirection(direction);
  image->FillBuffer(0.0);
  
  int qrows = 1;
  for (int j=0; j< Dimension; j++) qrows *= size[j];
  flann::Matrix<float> pixels(new float[qrows*cols],qrows,cols);
  

  typedef itk::ImageRegionIterator<OutputImageType> OutputImageRegionIteratorType;
  OutputImageRegionIteratorType phIt(image,image->GetLargestPossibleRegion());
  phIt.GoToBegin();
 
  typedef OutputImageType::IndexType IndexType;
  int k = 0;
  while(!phIt.IsAtEnd())
    {
      IndexType index;
      index = phIt.GetIndex();
      OutputImageType::PointType p;
      image->TransformIndexToPhysicalPoint(index,p);
      for (int j=0; j< Dimension; j++) pixels[k][j] = p[j];
      ++phIt, ++k;
    }
  std::cout << "\tfinding nearest neighbors to pixels ..." << std::endl;
  logfile << "\tfinding nearest neighbors to pixels ..." << std::endl;

  flann::Matrix<int> indices(new int[pixels.rows*nn], pixels.rows, nn); 
  flann::Matrix<float> dists(new float[pixels.rows*nn], pixels.rows, nn);
  // construct an randomized kd-tree index using 4 kd-trees 
  flann::Index<flann::L2<float> > index(points, flann::KDTreeIndexParams(4)); 
  index.buildIndex();
  // do a knn search, using 128 checks
  index.knnSearch(pixels, indices, dists, nn, flann::SearchParams(128));
 
  typedef itk::Image<float, Dimension>  OutputImageType;
  OutputImageType::Pointer distImage = OutputImageType::New();
  distImage->SetRegions(size);
  distImage->Allocate();
  distImage->SetOrigin(origin);
  distImage->SetSpacing(spacing );
  distImage->SetDirection(direction);
  distImage->FillBuffer(0.0);



  typedef itk::Image<float, Dimension>  OutputImageType;
  OutputImageType::Pointer speedImage = OutputImageType::New();
  speedImage->SetRegions(size);
  speedImage->Allocate();
  speedImage->SetOrigin(origin);
  speedImage->SetSpacing(spacing );
  speedImage->SetDirection(direction);
  speedImage->FillBuffer(0.0);

 // dist image
  typedef itk::ImageRegionIterator<OutputImageType> OutputImageRegionIteratorType;
  OutputImageRegionIteratorType distIt(distImage,distImage->GetLargestPossibleRegion());
 
  
  distIt.GoToBegin();
 

  //print them to a file
  
  std::ofstream myfile;
  if ( debug )
    myfile.open("dists.txt");

  
  for(int k =0; k < qrows; k++)
    { 
      for (int n=0; n < nn; n++) myfile << dists[k][n] << " ";
      if (debug)  myfile << std::endl;
      double d = vcl_sqrt(dists[k][0]);
      //double d = dists[k][0];
      if ( d < T ) //do svd in a narrowband
	{
	  std::vector< vnl_vector_fixed<float,Dimension>  > nps(nn);
	  vnl_vector_fixed<float,Dimension> qp;
	  for ( int j=0; j<Dimension; j++) qp[j] = pixels[k][j];
	  for (int n=0; n<nn; n++)
	    {
	      for ( int j=0; j<Dimension; j++) nps[n][j] = points[indices[k][n]][j];
	    }
	  d = computeDistance(nps, qp);
	  //speedIt.Set(0.0);
       }
      //else
      //	speedIt.Set(propWeight);

      distIt.Set(d);
      ++distIt;
      //++speedIt;
    }
  if (debug) myfile.close();
  if (debug)
    {
      writer->SetFileName("nnDist.nii");
      writer->SetInput(distImage);
      writer->Update();
    }
 
  std::cout << "\tcurvature filtering of distance image ..." << std::endl;
  logfile << "\tcurvature filtering of distance image ..." << std::endl;

  typedef itk::CurvatureFlowImageFilter<OutputImageType,OutputImageType> FilterType;
  FilterType::Pointer curveFilter = FilterType::New();
  curveFilter->SetInput(distImage);
  curveFilter->SetNumberOfIterations( cIter );
  curveFilter->SetTimeStep( 0.06 );
  curveFilter->UseImageSpacingOn();
  curveFilter->Update();
  OutputImageRegionIteratorType speedIt(speedImage,speedImage->GetLargestPossibleRegion());
  speedIt.GoToBegin();
  OutputImageRegionIteratorType fIt(curveFilter->GetOutput(), curveFilter->GetOutput()->GetLargestPossibleRegion());
  fIt.GoToBegin();
  while(!speedIt.IsAtEnd())
    {
      if ( fIt.Get() < T/2.0 )
	speedIt.Set( 0 );
      else
	speedIt.Set( propWeight );
      ++speedIt;
      ++fIt;
    }

  if (debug)
    {
      writer->SetFileName("filteredDist.nii");
      writer->SetInput(curveFilter->GetOutput());
      writer->Update();
    }
  if ( debug )
    {
      writer->SetFileName("speed.nii");
      writer->SetInput( speedImage );
      writer->Update();
    }
 
  typedef itk::BinaryThresholdImageFilter< OutputImageType,
                        OutputImageType    >       ThresholdingFilterType;
  ThresholdingFilterType::Pointer threshFilter = ThresholdingFilterType::New();
  threshFilter->SetInput(curveFilter->GetOutput());
  threshFilter->SetLowerThreshold( 0.0 );
  threshFilter->SetUpperThreshold( T );
  threshFilter->SetInsideValue(-0.5);
  threshFilter->SetOutsideValue(+0.5);
  
  
  typedef itk::ReinitializeLevelSetImageFilter<OutputImageType> 
    ReinitializeImageFilterType;
  ReinitializeImageFilterType::Pointer reinitFilter = ReinitializeImageFilterType::New();
  reinitFilter->NarrowBandingOff();
  reinitFilter->SetInput(threshFilter->GetOutput());
  reinitFilter->Update();
  if (debug)
    {
      writer->SetFileName("init.nii");
      writer->SetInput(reinitFilter->GetOutput());
      writer->Update();
    }

  typedef itk::BinaryMinMaxCurvatureFlowImageFilter<
    OutputImageType, OutputImageType > MinMaxFilterType;
  MinMaxFilterType::Pointer minMaxFilter = MinMaxFilterType::New();
  typedef MinMaxFilterType::RadiusValueType RadiusType;

  if ( mIter > 0 )
    {
      minMaxFilter->SetTimeStep( 0.1 );
      minMaxFilter->SetNumberOfIterations( mIter );
      minMaxFilter->SetStencilRadius( 0 );
      minMaxFilter->SetThreshold( 1e32 ); //always do min flow
      minMaxFilter->SetInput( reinitFilter->GetOutput());
      if ( debug )
	{
	  writer->SetFileName("holeFilledInit.nii");
	  writer->SetInput(minMaxFilter->GetOutput());
	  writer->Update();
	}
      reinitFilter->SetInput(minMaxFilter->GetOutput());
      reinitFilter->Update();
      std::cout << "\tMin flow no. elapsed iterations: " << minMaxFilter->GetElapsedIterations() << std::endl;
      logfile << "\tMin flow no. elapsed iterations: " << minMaxFilter->GetElapsedIterations() << std::endl;
 
      if ( debug )
	{
	  writer->SetFileName("reinit.nii");
	  writer->SetInput(reinitFilter->GetOutput());
	  writer->Update();
	}
 
    }

  
  std::cout << "\tintialization done ..." << std::endl;
  logfile << "\tintialization done ..." << std::endl;

 
  std::cout << "\tactive contour ..." << std::endl;
  logfile << "\tactive contour ..." << std::endl;
  


  typedef itk::GeodesicActiveContourWithPropagationLevelSetFunction<OutputImageType> GACFunctionType;
  GACFunctionType::Pointer gacFunction = GACFunctionType::New();


  // Geodesic Active Countour
  typedef  itk::GeodesicActiveContourLevelSetImageFilter< OutputImageType,
                OutputImageType >    GeodesicActiveContourFilterType;
  GeodesicActiveContourFilterType::Pointer geodesicActiveContour =
                                     GeodesicActiveContourFilterType::New(); 

  geodesicActiveContour->SetSegmentationFunction(gacFunction);	
  geodesicActiveContour->SetAutoGenerateSpeedAdvection(false);
  geodesicActiveContour->SetPropagationScaling( -1.0 );
  geodesicActiveContour->SetCurvatureScaling( curveWeight );
  geodesicActiveContour->SetAdvectionScaling( 1.0 );
 
  geodesicActiveContour->SetSpeedImage( speedImage );

  geodesicActiveContour->SetMaximumRMSError( 1e-3 );
  geodesicActiveContour->SetNumberOfIterations( gIter );
  geodesicActiveContour->SetInput( reinitFilter->GetOutput() );
  geodesicActiveContour->SetFeatureImage( curveFilter->GetOutput() );
  geodesicActiveContour->GenerateAdvectionImage();
  //geodesicActiveContour->SetNumberOfLayers(20);

  try
    {
    geodesicActiveContour->Update();
    }
  catch( itk::ExceptionObject & excep )
    {
    std::cerr << "Exception caught in geodesic active filter!" << std::endl;
    logfile << "Exception caught in geodesic active filter!" << std::endl;
    std::cerr << excep << std::endl;
    } 

  std::cout << std::endl;
  std::cout << "\tmax. no. iterations: " << geodesicActiveContour->GetNumberOfIterations() << std::endl;
  std::cout << "\tmax. RMS error: " << geodesicActiveContour->GetMaximumRMSError() << std::endl;
  std::cout << std::endl;
  std::cout << "\tno. elpased iterations: " << geodesicActiveContour->GetElapsedIterations() << std::endl;
  std::cout << "\tRMS change: " << geodesicActiveContour->GetRMSChange() << std::endl;
  std::cout << "\n\tfinished." << std::endl;
  logfile << std::endl;
  logfile << "\tmax. no. iterations: " << geodesicActiveContour->GetNumberOfIterations() << std::endl;
  logfile << "\tmax. RMS error: " << geodesicActiveContour->GetMaximumRMSError() << std::endl;
  logfile << std::endl;
  logfile << "\tno. elpased iterations: " << geodesicActiveContour->GetElapsedIterations() << std::endl;
  logfile << "\tRMS change: " << geodesicActiveContour->GetRMSChange() << std::endl;
  logfile << "\n\tfinished." << std::endl;

  


  if ( !oNameProvided )
    {
      std::string str= basename(inputMeshName);
      unsigned found = str.find_last_of(".");
      char bname[1000];
      sprintf(bname,"%s",(str.substr(0,found)).c_str());
      char currdir[1000];
      
      char * unused=getcwd(currdir,1000);
      sprintf(outputImageName,"%s/%s%s",currdir,bname,".nii");
    }

  logfile.close();

 
  
  writer->SetFileName(outputImageName);
  writer->SetInput(geodesicActiveContour->GetOutput());
  writer->Update();


  delete[] points.ptr(); 
  delete[] pixels.ptr(); 
  delete[] indices.ptr(); 
  delete[] dists.ptr();

 
  return EXIT_SUCCESS;
}


 /*
  std::cout << "fast marching for intialization ..." << std::endl;

  // fast marching for initialization
  typedef  itk::FastMarchingImageFilter< OutputImageType, OutputImageType >    FastMarchingFilterType;
  FastMarchingFilterType::Pointer  fastMarching = FastMarchingFilterType::New();
  typedef FastMarchingFilterType::NodeContainer  NodeContainer;
  typedef FastMarchingFilterType::NodeType       NodeType;

  NodeContainer::Pointer seeds = NodeContainer::New();

  OutputImageType::IndexType  seedPosition;
  
  for (int i =0; i< Dimension; i++)
    seedPosition[i] = int(size[i]/2);
  // double minR = 1e32;
  //for (int i =0; i< Dimension; i++)
  //  if (size[i] < minR) minR = size[i];
  //const double initialDistance = minR/2.0;
  
  const double initialDistance = pow(qrows,1/3.0)/2.0;
  const double seedValue = - initialDistance;
  NodeType node;
  node.SetValue( seedValue );
  node.SetIndex( seedPosition );

  seeds->Initialize();
  seeds->InsertElement( 0, node );
  fastMarching->SetTrialPoints(  seeds  );
  fastMarching->SetOutputSize(curveFilter->GetOutput()->GetBufferedRegion().GetSize() );
  fastMarching->SetOutputOrigin(curveFilter->GetOutput()->GetOrigin());
  fastMarching->SetOutputSpacing(curveFilter->GetOutput()->GetSpacing());
  fastMarching->SetOutputDirection(curveFilter->GetOutput()->GetDirection());
 
  writer->SetFileName("fastMarching.nii");
  writer->SetInput(fastMarching->GetOutput());
  writer->Update();
*/
  
