/*
Author: Ali Gooya <ali dot gooya at gmail dot com>
*/
#include "TypeDefinitions.h"
#include "math.h"
#include <iostream>
#include <fstream>
#include <time.h>
#include <unistd.h>
#include <libgen.h>

double  ProjectShape(vnl_vector<double> *shape, int nb_modes, vnl_vector<double> *xBar, vnl_vector<double> *eigenValues, 
		     vnl_matrix<double> *modesMatrix, vnl_vector<double> *alignedShape, vnl_vector<double> *b)
{
  
  *b = (*modesMatrix)*( *shape - *xBar);
  
  for ( int i = 0; i < nb_modes; i++)
    {
      if ( (*b)[i] <= -3.0 * sqrt((*eigenValues)[i]))
	(*b)[i] = -3.0 * sqrt((*eigenValues)[i]);
      else if ( 3.0 * sqrt((*eigenValues)[i]) <= (*b)[i] )
	(*b)[i] = 3.0 * sqrt((*eigenValues)[i]);
    }
  
  *alignedShape = (*modesMatrix).transpose()*(*b) + *xBar;
  vnl_vector<double> dist = *alignedShape - *shape;
  double d = dist.two_norm()/((double)dist.size());
  return d;
}

void print_help()
{
  std::cerr << std::endl;
  std::cerr << "Projects a series of VIRTUALY algined meshfiles to a PCA model." << std::endl;
  std::cerr << "Usage: projectVirtuallyAlignedShapesToModel " << std::endl;   
  std::cerr << "-n number of input meshes " << std::endl;
  std::cerr << "-i input mesh list"  << std::endl;
  std::cerr << "-m number of modes" << std::endl;
  std::cerr << "-p input PCA file name " << std::endl;
  std::cerr << "Options: " << std::endl;
  std::cerr << " -g PCA projected generic mesh file name (default: pcaPrj)" << std::endl;
  std::cerr << " -b output b vectors file name (default: bV.txt)" << std::endl;
  std::cerr << " -o PCA projected generic mesh file name list (default: prjList.txt)" << std::endl;
  std::cerr << " -d set if distances of projected to original meshes are desired (default: 0)" << std::endl;
  std::cerr << " -D output file name listing distances (default: dist.txt) " << std::endl;
  std::cerr << std::endl;

}




int main( int argc, char *argv[] )
{
  
  int c;
  unsigned int nb_mesh;
  char inpMeshList[1000];
  char pcaFileName[1000];
  char bFileName[1000]= "bV.txt";
  unsigned int nb_Modes;
  unsigned int estD=0;
  char distFileName[1000]="dist.txt";
  char outMeshList[1000]="prjList.txt";
  char prjGenName[1000]="pcaPrj";
  int rOpts = 0;
  while ((c = getopt (argc, argv, "n:i:p:m:b:d:D:o:g:")) != -1)
    switch (c)
      {
      case 'n':
	nb_mesh = atoi(optarg);
	++rOpts;
	break;
      case 'i':
	sprintf(inpMeshList,"%s",optarg);
	++rOpts;
	break;
      case 'p':
	sprintf(pcaFileName,"%s",optarg);
	++rOpts;
	break;
      case 'm':
	nb_Modes = atoi(optarg);
	++rOpts;
	break;
      case 'b':
	sprintf(bFileName,"%s",optarg);
	break;
      case 'd':
	estD = atoi(optarg);
	break;
      case 'D':
	sprintf(distFileName,"%s",optarg);
	break;
      case 'o':
	sprintf(outMeshList,"%s",optarg);
	break;
      case 'g':
	sprintf(prjGenName,"%s",optarg);
	break;
     case '?':
	fprintf (stderr, "Unknown option %s.\n", optarg);
	return 1;
      default:
	abort ();
      }
  if (rOpts != 4 )
    {
      print_help();
      return 1;
    }
        

  std::fstream meshNames;
  meshNames.open(inpMeshList);

  //Next, we make pointset
  //Something to keep the point sets
  std::vector< vnl_vector<double> *> shapesVector( nb_mesh );
  std::vector<char *> meshNamesVector(nb_mesh);
  for ( unsigned int k = 0; k < nb_mesh; k++ )
    {
      meshNamesVector[k] = new char[1000];
    }

  for ( unsigned int k = 0; k < nb_mesh; k++ )
    {
      std::fstream myfile;
      char mesh[1000];
      meshNames >> meshNamesVector[k];
      myfile.open(meshNamesVector[k]);
      vnl_vector<double> *points = new vnl_vector<double>;
      shapesVector[k] = points;
      //     while ( !myfile.eof() )
      myfile >> (*shapesVector[k]);
      //std::cout << *shapesVector[k] << std::endl;            
      myfile.close();
    }

  //read model
  std::fstream pcaModel;
  pcaModel.open(pcaFileName);

  //int nb_Modes = atoi(argv[3]);  
  int n;
  pcaModel >> n;

  vnl_vector<double> xBar((n-1)*GMMPointDimension);
  pcaModel >> xBar;
  //std::cout << "xBar: " << xBar << std::endl;

  vnl_vector<double> eigenValues(nb_Modes);
  pcaModel >> eigenValues ;
  pcaModel.ignore(1000000, '\n');
  //std::cout << "eigenValues: " << eigenValues << std::endl;
  




  vnl_matrix<double> modesMatrix(nb_Modes, (n-1)*GMMPointDimension);
  pcaModel >> modesMatrix;
  //std::cout << modesMatrix << std::endl;
  pcaModel.close();



  std::vector< vnl_vector<double> *> alignedShapesVector( nb_mesh );

  for ( unsigned int k = 0; k < nb_mesh; k++ )
    {
      vnl_vector<double> *points = new vnl_vector<double>;
      alignedShapesVector[k] = points;
    }

  std::vector< vnl_vector<double> >  bVectors(nb_mesh);
  std::ofstream bfile;
  bfile.open(bFileName);

  //now projection
  double *d = new double[nb_mesh]; 

  std::ofstream distfile;
  if (estD)
    distfile.open(distFileName);
  
  std::ofstream ofileList;
  ofileList.open(outMeshList);

  for ( unsigned int k = 0; k < nb_mesh; k++ )
    {
      d[k]=ProjectShape(shapesVector[k], nb_Modes, &xBar, &eigenValues, &modesMatrix, alignedShapesVector[k], &bVectors[k]); 
      if (estD)
	distfile << d[k] << " ";
      //char filename[1000];
      //sprintf(name,"%s%02d%s", argv[4],k+1,".txt");
      std::string str= basename(meshNamesVector[k]);
      unsigned found = str.find_last_of(".");
      char bname[1000];
      sprintf(bname,"%s",(str.substr(0,found)).c_str());
      char filename[1000];
      char currdir[1000];
      
      char * unused=getcwd(currdir,1000);
      sprintf(filename,"%s/%s_%s%s",currdir,bname,prjGenName,".txt");
      ofileList << filename << std::endl;


      std::ofstream myfile;
      myfile.open(filename);
      for (int i =0; i< n-1 ; i++)
	{
	  for ( int q=0; q< GMMPointDimension; q++)
	    myfile << (*alignedShapesVector[k])[i*GMMPointDimension+q] << " ";
	  myfile << std::endl;
	}
      bfile << bVectors[k] << std::endl;
      myfile.close();
    }
  if (estD)
    distfile.close();
  bfile.close();
  ofileList.close();




  for ( unsigned int k = 0; k < nb_mesh; k++ )
    {
      delete shapesVector[k];
    }
  for ( unsigned int k = 0; k < nb_mesh; k++ )
    {
      delete alignedShapesVector[k];
    }
  delete d;

  
  return EXIT_SUCCESS;
}


