/*
Author: Ali Gooya <ali dot gooya at gmail dot com>
*/
#include "stdafx.h"
#include "math.h"
#include <iostream>
#include <fstream>
#include <time.h>
#include <unistd.h>
#include <libgen.h>

#include "interpolation.h"
//#include "vnl/vnl_sample.h"
#include <vnl/vnl_vector_fixed.h>



using namespace alglib;

void print_help()
{
  std::cerr << std::endl;
  std::cerr << "Interpolates a vector field on a point set and (optionally) warps it." << std::endl;
  std::cerr << "Usage: warpPointSetUsingRBF " << std::endl;
  std::cerr << " -q queryPointSet \ttext file including space separated coordinates of each query point per line." << std::endl;
  std::cerr << " -i inputPointSet \ttext file including space separated coordinates of input point per line." << std::endl;
  std::cerr << " -f inputVectorField \ttext file including observed vectors at input points." << std::endl;

  std::cerr << "Options:" << std::endl;
  std::cerr << " -w set if warping is needed (default: 0) " << std::endl;
  std::cerr << " -o outputVectorField/PointSet (default: <inputVectorField>_inerp.txt/<queryPointSet>_warped.txt) " << std::endl;
  std::cerr << " -s scale for RBF interpolation (default: 4.0)" << std::endl;
  std::cerr << std::endl;

}

#define Dimension 3

int main( int argc, char *argv[] )
{
 
  int rOpts = 0;
  int c;
  char queryPointSetFileName[1000];
  char inputPointSetFileName[1000];
  char inputVectorFieldFileName[1000];
  char outputPointSetFileName[1000];
  float scale = 4.0;
  int oNameProvided = 0;
  int warp = 0;


  while ((c = getopt (argc, argv, "q:i:f:o:s:w:")) != -1)
    switch (c)
      {
      case 'q':
	sprintf(queryPointSetFileName,"%s",optarg);
	++rOpts;
	break;
      case 'i':
	sprintf(inputPointSetFileName,"%s",optarg);
	++rOpts;
	break;
      case 'f':
	sprintf(inputVectorFieldFileName,"%s",optarg);	
	++rOpts;
	break;
      case 's':
	scale = atof(optarg);
	break;
      case 'o':
	sprintf(outputPointSetFileName,"%s",optarg);
	oNameProvided = 1;
	break;
      case 'w':
	warp = atoi(optarg);
	break;
      case '?':
	fprintf (stderr, "Unknown option %s.\n", optarg);
	return 1;
      default:
	abort ();
      }
  
 

  if (rOpts != 3 )
    {
      print_help();
      return 1;
    }

  typedef vnl_vector_fixed < float, Dimension > PointType;
  typedef std::vector< PointType > PointSetType;
  
  typedef vnl_vector_fixed < float, Dimension > VectorType;
  typedef std::vector< VectorType > VectorFieldType;


  std::fstream inputPointSetFile;
  inputPointSetFile.open(inputPointSetFileName);
  PointSetType inputPointSet;

  std::fstream inputVectorFieldFile;
  inputVectorFieldFile.open(inputVectorFieldFileName);
  VectorFieldType inputVectorField;

  int iCnt = 0; 
  PointType p;
  VectorType v; 
  while( (inputPointSetFile >> p) && (inputVectorFieldFile >> v) )
    {
 
      //inputPointSetFile >> p;
      //inputVectorFieldFile >> v;
      inputPointSet.push_back(p);
      inputVectorField.push_back(v);
      ++iCnt;
      inputPointSetFile.ignore(1000000,'\n');
      inputVectorFieldFile.ignore(1000000,'\n');
     }
  inputPointSetFile.close();
  inputVectorFieldFile.close();
  //iCnt -= 1;
  
  std::fstream queryPointSetFile;
  queryPointSetFile.open(queryPointSetFileName);
  PointSetType queryPointSet;

  int qCnt = 0;
  
  while(    queryPointSetFile >> p )
    {
      queryPointSet.push_back(p);
      ++qCnt;
      queryPointSetFile.ignore(1000000,'\n');
     }
  queryPointSetFile.close();
  
  //RBF stuff
  rbfmodel model;
  rbfcreate(Dimension, Dimension, model);
  real_2d_array xy;
  xy.setlength( iCnt, Dimension+Dimension);
  for (int i=0; i< iCnt; i++)
    {
      VectorType v = inputVectorField[i];
      PointType p = inputPointSet[i];
      for (int j=0; j< Dimension; j++)
	xy[i][j] = p[j];
      for (int j=0; j< Dimension; j++)
	xy[i][Dimension+j] = v[j];
    }

  rbfsetpoints(model, xy);
  rbfreport rep;
  rbfsetalgomultilayer(model,scale,2,1e-1);
  rbfbuildmodel(model, rep);

  if( int(rep.terminationtype)) // EXPECTED: 1
    std::cout << "making rbf model ..." << std::endl;
  else
    {
      std::cerr << "unable to make rbf model " << std::endl;
      return 1;
    }
 
  //PointSetType outputPointSet;
 // writting outputPointSet to a file
  if ( !oNameProvided )
    {
      if ( warp )
	{
	  std::string str= basename(queryPointSetFileName);
	  unsigned found = str.find_last_of(".");
	  char bname[1000];
	  sprintf(bname,"%s",(str.substr(0,found)).c_str());
	  char currdir[1000];
      
	  char * unused=getcwd(currdir,1000);
	  sprintf(outputPointSetFileName,"%s/%s%s",currdir,bname,"_warped.txt");
	}
      else
	{
	  std::string str= basename(inputVectorFieldFileName);
	  unsigned found = str.find_last_of(".");
	  char bname[1000];
	  sprintf(bname,"%s",(str.substr(0,found)).c_str());
	  char currdir[1000];
      
	  char * unused=getcwd(currdir,1000);
	  sprintf(outputPointSetFileName,"%s/%s%s",currdir,bname,"_interp.txt");
	}
    }
 
  
  
  std::ofstream outputPointSetFile;
  outputPointSetFile.open(outputPointSetFileName);
  //now querying at each point
  std::cout << "size of query point set: " << queryPointSet.size() <<  std::endl;
   
  for(int i=0; i< qCnt; i++)
    {
      real_1d_array y;
      real_1d_array x;
      x.setlength(Dimension);
      y.setlength(Dimension);
      for (int j=0; j<Dimension; j++) x[j] = queryPointSet[i][j];
      rbfcalc(model,x,y);
      VectorType dp;
      if (warp)
	for (int j=0; j<Dimension; j++) dp[j] = x[j] + y[j];
      else
 	for (int j=0; j<Dimension; j++) dp[j] = y[j];

      if ( i < qCnt -1 )
	outputPointSetFile << dp << std::endl;
      else
	outputPointSetFile << dp;
    }
  
 
  xy.setlength(0,0);
  return EXIT_SUCCESS;
}


